#!/usr/bin/env python3

if __name__ == "__main__":
    import argparse
    import getpass
    import os
    import sys

    import requests

    from .docker import DispatchDocker
    from .remote import DispatchRemote
    from . import DispatchBase
    from .model import load_model
    from .constants import DispatchStage

    ## Remote the pre-processing from the dispatchers
    class DispatchRemoteSimple(DispatchRemote):
        def _run_processing(self, inp):
            return

    class DispatchDockerSimple(DispatchDocker):
        def _run_processing(self, inp):
            return

    ## First parser, this handles 
    parser = argparse.ArgumentParser(description="A CLI for running inferences on ABLInfer servers or Docker containers with an ABLInfer model specification")
    parser.add_argument("-s", "--server", help="Base URL of the remote ABLInfer server to use, instead of a local Docker instance")
    parser.add_argument("-c", "--cert", type=str, default=False, help="SSL certificate to use to verify the server; if none, SSL verification will be skipped")
    parser.add_argument("-u", "--username", type=str, help="Username for remote server")
    parser.add_argument("-p", "--password", type=str, help="Password for remote server (if none and username is provided, the password will be prompted)")
    parser.add_argument("-f", "--field-names", action="store_true", help="Use field names for model arguments instead of the model flags (only affects this CLI interface, no effect on the actual inference)")
    parser.add_argument("-w", "--suppress-verify", action="store_true", help="Suppress the SSL verification warning (e.g. for self-signed certificates")
    parser.add_argument("model", help="Either a local model specification file or a model ID from the server")
    parser.add_argument("args", nargs=argparse.REMAINDER, help="Model arguments")

    args = parser.parse_args()

    dispatch: DispatchBase

    if args.suppress_verify:
        import urllib3
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

    if args.server is not None: ## Remote dispatch
        session = requests.Session()
        session.verify = args.cert
        if args.username is not None:
            if args.password is None:
                args.password = getpass.getpass("Password: ")
            session.auth = (args.username, args.password)

        dispatch = DispatchRemoteSimple({
            "base_url": args.server,
            "session": session,
        })
        
        if not os.path.isfile(args.model):
            try:
                model = dispatch.get_model(args.model)
            except Exception as e:
                sys.stderr.write("Unable to retrieve model \"%s\" from server: %s" % (args.model, repr(e)))
                sys.stderr.write("Available models on server: %s" % ", ".join(dispatch.get_model_list()))
                sys.exit(1)
        else:
            with open(args.model, 'r') as f:
                model = load_model(f)[0]
    else: ## Local dispatch
        dispatch = DispatchDockerSimple()

        with open(args.model, 'r') as f:
            model = load_model(f)[0]

    ## Assemble the second parser
    flag_map = {}
    sec_map = {}
    for s in ("outputs", "inputs", "params"):
        for k, v in model[s].items():
            sec_map[k] = s
            if args.field_names:
                flag_map[k] = "--" + k.replace('_', '-')
            else:
                flag_map[k] = v["flag"].rstrip('=')
        
    if not model.get("order"):
        order = sorted(flag_map)
    else:
        order = model["order"]

    parser = argparse.ArgumentParser(args.model)
    for k in order:
        v = model[sec_map[k]][k]
        
        flag = flag_map[k]
        params = {
            "help": "%s. %s" % (v["name"], v["description"]),
            "dest": k,
        }

        typ = v["type"]
        if typ in ("segmentation", "volume"): ## File
            params["required"] = v["status"] == "required"
        elif typ == "int":
            params["type"] = int
        elif typ == "float":
            params["type"] = float
        elif typ == "bool":
            params["action"] = "store_true"
        elif typ == "enum":
            params["choices"] = tuple(k["enum"].values())
        elif typ == "string":
            pass
        else:
            raise ValueError("Unknown type %s" % typ)

        parser.add_argument(flag, **params)

    args = parser.parse_args(args.args)
    model_config = {
        "inputs": {},
        "params": {},
        "outputs": {},
    }
    for k in order:
        s = sec_map[k]
        if getattr(args, k) is not None:
            model_config[s][k] = {"value": getattr(args, k)}
            if s in ("inputs", "outputs"):
                model_config[s][k]["enabled"] = True
        elif s in ("inputs", "outputs"):
            model_config[s][k] = {"enabled": False}

    def progress(stage, f1, f2, s):
        if stage == DispatchStage.Initial:
            stage = "INIT"
        elif stage == DispatchStage.Validate:
            stage = "VALI"
        elif stage == DispatchStage.Preprocess:
            stage = "PREP"
        elif stage == DispatchStage.Save:
            stage = "SAVE"
        elif stage == DispatchStage.Run:
            stage = "RUN "
        elif stage == DispatchStage.Load:
            stage = "LOAD"
        elif stage == DispatchStage.Postprocess:
            stage = "POST"
        else:
            stage = "    "

        if s.endswith('\n'):
            s = s[:-1]
        print("[%s] %3.0f%%, %3.0f%%: %s" % (stage, 100*f1, 100*f2, s))

    dispatch.run(
        model,
        model_config,
        progress=progress,
    )



import collections
import typing
from typing import Generic

if hasattr(typing, "get_args"):
    from typing import get_args, get_origin
else:
    ## https://stackoverflow.com/questions/49171189/whats-the-correct-way-to-check-if-an-object-is-a-typing-generic
    if hasattr(typing, '_GenericAlias'):
        # python 3.7
        def _is_generic(cls):
            if isinstance(cls, typing._GenericAlias):
                return True

            if isinstance(cls, typing._SpecialForm):
                return cls not in {typing.Any}

            return False


        def _is_base_generic(cls):
            if isinstance(cls, typing._GenericAlias):
                if cls.__origin__ in {typing.Generic, typing._Protocol}:
                    return False

                if isinstance(cls, typing._VariadicGenericAlias):
                    return True

                return len(cls.__parameters__) > 0

            if isinstance(cls, typing._SpecialForm):
                return cls._name in {'ClassVar', 'Union', 'Optional'}

            return False
    else:
        # python <3.7
        if hasattr(typing, '_Union'):
            # python 3.6
            def _is_generic(cls):
                if isinstance(cls, (typing.GenericMeta, typing._Union, typing._Optional, typing._ClassVar)):
                    return True

                return False


            def _is_base_generic(cls):
                if isinstance(cls, (typing.GenericMeta, typing._Union)):
                    return cls.__args__ in {None, ()}

                if isinstance(cls, typing._Optional):
                    return True

                return False
        else:
            # python 3.5
            def _is_generic(cls):
                if isinstance(cls, (typing.GenericMeta, typing.UnionMeta, typing.OptionalMeta, typing.CallableMeta, typing.TupleMeta)):
                    return True

                return False


            def _is_base_generic(cls):
                if isinstance(cls, typing.GenericMeta):
                    return all(isinstance(arg, typing.TypeVar) for arg in cls.__parameters__)

                if isinstance(cls, typing.UnionMeta):
                    return cls.__union_params__ is None

                if isinstance(cls, typing.TupleMeta):
                    return cls.__tuple_params__ is None

                if isinstance(cls, typing.CallableMeta):
                    return cls.__args__ is None

                if isinstance(cls, typing.OptionalMeta):
                    return True

                return False


    def is_generic(cls):
        """
        Detects any kind of generic, for example `List` or `List[int]`. This includes "special" types like
        Union and Tuple - anything that's subscriptable, basically.
        """
        return _is_generic(cls)


    def is_base_generic(cls):
        """
        Detects generic base classes, for example `List` (but not `List[int]`)
        """
        return _is_base_generic(cls)


    def is_qualified_generic(cls):
        """
        Detects generics with arguments, for example `List[int]` (but not `List`)
        """
        return is_generic(cls) and not is_base_generic(cls)

    def get_args(tp):
        """Get type arguments with all substitutions performed.
        For unions, basic simplifications used by Union constructor are performed.
        Examples::
            get_args(Dict[str, int]) == (str, int)
            get_args(int) == ()
            get_args(Union[int, Union[T, int], str][int]) == (int, str)
            get_args(Union[int, Tuple[T, int]][str]) == (int, Tuple[str, int])
            get_args(Callable[[], T][int]) == ([], int)
        """
        if is_qualified_generic(tp):
            res = tp.__args__
            if get_origin(tp) is collections.abc.Callable and res[0] is not Ellipsis:
                res = (list(res[:-1]), res[-1])
            return res
        return ()

    def get_origin(tp):
        """Get the unsubscripted version of a type.
        This supports generic types, Callable, Tuple, Union, Literal, Final and ClassVar.
        Return None for unsupported types. Examples::
            get_origin(Literal[42]) is Literal
            get_origin(int) is None
            get_origin(ClassVar[int]) is ClassVar
            get_origin(Generic) is Generic
            get_origin(Generic[T]) is Generic
            get_origin(Union[T, int]) is Union
            get_origin(List[Tuple[T, T]][int]) == list
        """
        if is_generic(tp):
            return tp.__origin__
        if tp is Generic:
            return Generic
        return None



#!/usr/bin/env python3

from collections import OrderedDict as OD
import logging
import os
from urllib.parse import urljoin
import time
from typing import List

import requests as r

from ..base import DispatchBase, DispatchException
from ..constants import DispatchStage
from .util import ProgressReporter, save_resp, FObjReadWrapper

def urljoin_b(*args):
    """Make urljoin behave like path.join to preserve my sanity."""
    if len(args) > 2:
        return urljoin(urljoin_b(*args[:-1])+'/', args[-1])
    elif len(args) == 2:
        return urljoin(args[0]+'/', args[1])
    return args[0]

class DispatchRemote(DispatchBase):
    """Class for dispatching to an ABLInfer server.

    A required ``base_url`` key is added to ``config``, which must be the server's base URL, which
    will be passed to :func:`urllib.parse.urljoin` to construct the query URLs. In addition, a
    ``session`` parameter is added to ``config`` which allows the user to provide a 
    :class:`requests.Session` instance for SSL verification or authentication. An
    ``ignore_mismatch`` parameter has been added to config; if ``True``, any mismatch between the 
    local version and the server version will be ignored.
    """
    def __init__(self, config=None):
        self.base_url = None
        self.session = None
        self.remote_session = None
        self.model_id = None
        self.ignore_mismatch = False

        super().__init__(config=config)

    def get_model_list(self) -> List[str]:
        """Retrieve the list of models available on the site."""
        with self._lock:
            resp = self.session.get(urljoin_b(self.base_url, "models"))
            resp.raise_for_status()
            return resp.json()["data"]

    def get_model(self, model_id: str):
        """Retrieve a model from the site.

        This function assumes that the model received is normalized.

        :param model_id: The model's ID.
        """
        with self._lock:
            resp = self.session.get(urljoin_b(self.base_url, "models", model_id))
            resp.raise_for_status()
            return resp.json(object_pairs_hook=OD)["data"]

    def _validate_config(self):
        super()._validate_config()

        self.base_url = self.config["base_url"]
        if not self.base_url.endswith('/'):
            self.base_url += '/'
        self.session = self.config["session"] if "session" in self.config else r.Session()
        self.ignore_mismatch = self.config.get("ignore_mismatch", False)

        ## Check the server
        logging.info("Trying server at %s..." % self.base_url)
        resp = self.session.get(self.base_url)
        resp.raise_for_status()
        resp = resp.json()
        if resp["data"]["server"] != "inferserver":
            raise ValueError("Unknown server %s" % repr(resp["data"]["server"]))

    def _validate_model_config(self):
        ## We need to check that the server has the correct version of the model first
        self.model_id = self.model["id"]
        try:
            model = self.get_model(self.model_id)
        except Exception as e:
            raise DispatchException("Unable to retrieve model from the server: %s" % repr(e))

        if not self.ignore_mismatch and self.model["version"] != model["version"]:
            raise DispatchException("Version mismatch between server model and local model: server has v%s, we have v%s" % (model["version"], self.model["version"]))

        super()._validate_model_config()

    def _make_fmap(self):
        return {}

    def _make_flags(self, fmap):
        return []

    def _make_command(self, flags):
        resp = self.session.post(
            urljoin_b(self.base_url, "models", self.model_id), 
            json={
                "inputs": {n: {"enabled": v["enabled"]} for n, v in self.model_config["inputs"].items()},
                "params": self.model_config["params"],
                "outputs": {n: {"enabled": v["enabled"]} for n, v in self.model_config["outputs"].items()},
            },
        )
        resp.raise_for_status()

        self.remote_session = resp.json()["data"]["session_id"]
        self.progress(DispatchStage.Validate, 0.5, 1, "Session ID is %s" % (self.remote_session))

        return []

    def _get_status(self):
        return self.session.get(urljoin_b(self.base_url, "sessions", self.remote_session)).json()["data"]["status"]

    def _save_input(self, fmap):
        total = len(self.model_config["inputs"])
        for n, (name, v) in enumerate(self.model_config["inputs"].items()):
            if not v["enabled"]:
                logging.info("Skipping disabled input %s" % name)
                continue
            string = "Uploading %s..." % name
            logging.info(string)
            with open(v["value"], "rb") as f:
                header = f.read(1024)
                f.seek(0)
                resp = self.session.post(urljoin_b(self.base_url, "models", self.model["id"], "inputs", name, "check"), data=header)
                resp.raise_for_status()
                j = resp.json()
                if not j["data"]["acceptable"]:
                    ft = j["data"]["acceptable"]
                    ft = ft if ft is not None else "unknown"
                    raise DispatchException("Invalid filetype for input %s; expected %s, got %s" % (name, self.model["inputs"][name]["extension"], j["data"]["filetype"]))

                fwrap = FObjReadWrapper(f, os.path.getsize(v["value"]), string, lambda f, s: self.progress(DispatchStage.Save, n/total + f/total, f, s), period=0.1)
                resp = self.session.put(urljoin_b(self.base_url, "sessions", self.remote_session, "inputs", name), headers={"Content-Type": "application/octet-stream"}, data=fwrap)
                resp.raise_for_status()
        if self._get_status() == "waiting":
            raise DispatchException("Session ID %s is still waiting for input, but all input has been provided, please report this" % self.remote_session)

    def _run_command(self, cmd):
        logging.info("Starting run...")
        resp = self.session.get(urljoin_b(self.base_url, "sessions", self.remote_session, "logs"), stream=True)
        for line in resp.iter_lines(5):
            if line == b'\0':
                continue
            self.progress(DispatchStage.Run, 0, 0, line.decode("utf-8"))

        ## Now the run is over
        logging.info("Logs ended, waiting for the session to finish...")
        while True:
            status = self._get_status()
            if status in ("complete", "failed"):
                break
            time.sleep(1)

        if status == "failed":
            raise DispatchException("Session ID %s failed, please report this" % self.remote_session)

    def _load_output(self, fmap):
        total = len(self.model_config["outputs"].items())
        for n, (name, v) in enumerate(self.model_config["outputs"].items()):
            if not v["enabled"]:
                logging.info("Skipping disabled output %s" % name)
                continue
            logging.info("Saving output %s" % name)
            self._output_files.append(v["value"])
            resp = self.session.get(urljoin_b(self.base_url, "sessions", self.remote_session, "outputs", name), stream=True)
            resp.raise_for_status()
            save_resp(resp, v["value"], "Saving output %s..." % name, lambda f, s: self.progress(DispatchStage.Load, n/total + f/total, f, s), period=0.1)

    def _cleanup(self, error=None):
        super()._cleanup(error=error)

        self.remote_session = None
        self.session = None
        self.model_id = None



#!/usr/bin/env python3

import logging
from typing import Mapping

def normalize_model_config(model: Mapping, model_config: Mapping) -> Mapping:
    """Normalizes the given model config against the given model.

    The model is assumed to be normalized.

    Note that for processing configuration to be normalized properly, the processing functions must 
    be imported before running this function; e.g. :mod:`ablinfer.slicer.processing` requires 
    dependencies not usually found outside 3DSlicer and so is not imported automatically.

    :param model: The model.
    :param model_config: The model configuration to normalize; this will modify it.
    :returns: The modified model config, for convenience.
    """

    for (s, process) in (("inputs", "pre"), ("outputs", "post")):
        csec = model_config[s]
        for name, spec in model[s].items():
            if name not in csec:
                raise ValueError("Missing %s %s" % (s.rstrip('s'), name))
            sec = csec[name]
            if "enabled" not in sec:
                sec["enabled"] = spec["status"] in ("required", "suggested")
            elif not sec["enabled"] and spec["status"] == "required":
                raise ValueError("%s %s cannot be disabled" % (s.title()[:-1], name))

            if not sec["enabled"]:
                continue

            if "value" not in sec:
                raise ValueError("Missing value for %s %s" % (s.rstrip('s'), name))

            if process not in sec:
                sec[process] = [{} for i in range(len(spec[process]))]
            elif sec[process] and len(sec[process]) != len(spec[process]):
                raise ValueError("Either no processing parameters may be given or all must be given in %s/%s" % (s, name))
            
            for p, elem in zip(spec[process], sec[process]):
                defaults = {
                    "enabled": p["status"] in ("required", "suggested"),
                    "params": p["params"].copy()
                }
                if "enabled" not in elem:
                    elem["enabled"] = defaults["enabled"]
                elif p["status"] == "required" and not elem["enabled"]:
                    raise ValueError("Processing operation \"%s\" on %s/%s may not be disabled" % (p["name"], s, name))

                if "params" not in elem:
                    elem["params"] = defaults["params"]
                else:
                    for p, v in defaults["params"].items():
                        if p not in elem["params"]:
                            elem["params"][p] = v

    if "params" not in model_config:
        logging.warning("Missing params section")
        model_config["params"] = {}

    csec = model_config["params"]
    for name, spec in model["params"].items():
        if name not in csec:
            logging.warning("Using default value for params/%s" % name)
            csec[name] = {"value": spec["default"]}
            continue
        sec = csec[name]
        if "value" not in sec:
            sec["value"] = spec["default"]
            continue

        typ = spec["type"]
        if typ in ("int", "float"):
            try:
                if typ == "int":
                    sec["value"] = int(sec["value"])
                else:
                    sec["value"] = float(sec["value"])
            except Exception as e:
                raise ValueError("Invalid value for params/%s: %s" % (name, repr(e)))
            if sec["value"] < spec["min"] or sec["value"] > spec["max"]:
                raise ValueError("Value for params/%s must be on [%d, %d]" % (name, spec["min"], spec["max"]))
        elif typ == "string":
            try:
                sec["value"] = str(sec["value"])
            except Exception as e:
                raise ValueError("Invalid value for params/%s: %s" % (name, repr(e)))
        elif typ == "bool":
            try:
                sec["value"] = bool(sec["value"])
            except Exception as e:
                raise ValueError("Invalid value for params/%s: %s" % (name, repr(e)))
        elif typ == "enum":
            if sec["value"] not in spec["enum"].values():
                raise ValueError("Invalid enum value for params/%s" % name)
        else:
            raise NotImplementedError("Unknown type %s, either the model is invalid for this version or this is a bug" % repr(typ))

    return model_config



"""This module contains various constants."""

import enum

class DispatchStage(enum.Enum):
    Initial = 0
    Validate = 1
    Preprocess = 2
    Save = 3
    Run = 4
    Load = 5
    Postprocess = 6



"""Base class for dispatch helpers."""

from __future__ import division

from abc import ABCMeta, abstractmethod
import logging
import os
from threading import RLock
from typing import List, Mapping, Any, Callable, Dict

from .processing import dispatch_processing
from .model import normalize_model_config, update_normalize_model
from .constants import DispatchStage

class DispatchException(Exception):
    """To be raised if a process runs into a problem.

    :param stderr: The stderr output of the process, if any.
    """
    def __init__(self, msg, stderr: str = None):
        super().__init__(msg)

        self.stderr = stderr

class DispatchBase(metaclass=ABCMeta):
    """Base class for dispatching to models.

    This class cannot actually be instantiated; users are intended to instantiate a subclass that
    dispatches using the appropriate method (e.g. :class:`ablinfer.docker.DispatchDocker`).

    The general procedure when running is:

    1. :meth:`_validate_config`
    2. :meth:`_validate_model_config`
    3. :meth:`_make_fmap`
    4. :meth:`_make_flags`
    5. :meth:`_make_command`
    6. :meth:`_run_processing` (dispatches required nodes to :meth:`_clone` to copy them)
    7. :meth:`_save_input`
    8. :meth:`_run_command`
    9. :meth:`_load_output`
    10. :meth:`_run_processing`
    11. :meth:`_cleanup`

    Functions are expected to raise exceptions if they encounter any unfixable errors. If an
    exception is raised, :meth:`_cleanup_all` will be called.
    """

    def __init__(self, config: Mapping = None) ->  None:
        if config is None:
            config = {}
        self.config = config

        self._created_files: List[str] = []
        self._output_files: List[str] = []
        self._pre_nodes: Dict[str, Any] = {}

        self._lock = RLock()

        self.model: Mapping[str, Any] = None
        self.model_config: Mapping[str, Any] = None
        self.progress: Callable[[DispatchStage, float, float, str], None] = None

        self._validate_config()

    def _validate_config(self) ->  None:
        """Validate the non-model configuration."""

    def _validate_model_config(self) ->  None:
        """Validate the model configuration.

        This first normalizes the model config, which also validates it. It then ensures that no 
        inputs are null, no two outputs are the same, and no outputs are the same as inputs.

        Note that when overriding this method without delegating here, the new method **must** 
        normalize the model config by calling :func:``ablinfer.model.normalize_model_config``.
        """
        self.model_config = normalize_model_config(self.model, self.model_config)

        in_nodes = {i["value"] for i in self.model_config["inputs"].values() if i["enabled"]}
        if None in in_nodes:
            raise ValueError("Make sure that all inputs are selected!")

        out_nodes = {i["value"] for i in self.model_config["outputs"].values() if i["enabled"] and i["value"] is not None}
        if in_nodes.intersection(out_nodes).difference((None,)) or len(out_nodes) != sum((1 for i in self.model_config["outputs"].values() if i["enabled"] and i["value"] is not None)):
            raise ValueError("Inputs and outputs must be different")

    def _make_fmap_helper(self, actual_path: str) ->  Mapping[str, str]:
        ## Strip the trailing slash of actual_path, if it has one
        if len(actual_path) > 1 and actual_path[-1] in ('/', '\\'):
            actual_path = actual_path[:-1]

        fmap = {}
        for s in ("inputs", "outputs",):
            for k, v in self.model[s].items():
                fmap[k] = actual_path + '/' + k + v["extension"]

        return fmap

    @abstractmethod
    def _make_fmap(self) ->  Mapping[str, str]:
        """Make the file map.

        :returns: The created filemap, mapping the input/output name to its path as seen by
                  whatever the command will be dispatched to (e.g. for Docker, this would be the
                  path inside the container).
        """

    def _clone(self, node: Any) ->  Any:
        """Clone the given node for pre-processing.

        This should be overwritten for implementations based on the format your nodes are in. Any
        cloned node distinct from the input node is added to `self._pre_nodes`, which maps the
        input name to the result of this function. These nodes should be cleaned up at in the
        :meth:`_cleanup` function.
        """
        return node

    def _run_processing(self, inp: bool) ->  None:
        """Run pre/post processing.

        :param inp: Whether to do the input or output section.
        """
        section = "inputs" if inp else "outputs"
        process = "pre" if inp else "post"
        stage = DispatchStage.Preprocess if inp else DispatchStage.Postprocess

        if inp:
            self._pre_nodes = {}

        total = len(self.model[section])
        for count, (o, member) in enumerate(self.model[section].items()):
            node = self.model_config[section][o]["value"]

            if process not in member or not self.model_config[section][o]["enabled"]:
                continue
            string = "Running %sprocessing for %s..." % (process, member["name"])
            logging.info(string)
            self.progress(stage, count/total, 0, string)

            ## Clone nodes if we're pre-processing
            if inp and True in (i["enabled"] for i in self.model_config[section][o][process]):
                ## Make sure we have at least one pre-processing to run
                cnode = self._clone(node)
                if cnode != node:
                    self._pre_nodes[o] = cnode
                node = cnode

            total_ops = len(member[process])
            for n, op in enumerate(member[process], 1):
                logging.info("%d. %s" % (n, op["name"]))
                self.progress(stage, count/total+(n-1)/(total*total_ops), (n-1)/total_ops, string+" %d. %s" % (n, op["name"]))
                if not self.model_config[section][o][process][n-1]["enabled"]:
                    logging.info("- Not enabled: skipping")
                    continue
                ## TODO: Possibly keep going if a processing step fails?
                dispatch_processing(op, self.model_config[section][o][process][n-1], node, member, self.model, self.model_config, inp=inp)

    @abstractmethod
    def _save_input(self, fmap: Mapping[str, str]) ->  None:
        """Save the input files at the appropriate locations for running the model.

        See the :class:`ablinfer.docker.DispatchDocker` class for an example. This function will
        need to be overridden to reflect whatever node format your subclass uses.
        """

    def _make_flags(self, fmap: Mapping[str, str]) ->  List[str]:
        """Construct the flags to pass to the command.

        :param fmap: The generated fmap.
        """
        cmd = {}
        def format_flag(f, v):
            if not f:
                return (v,)
            elif f[-1] == '=':
                return (f+v,)
            return (f, v,)

        for n in ("inputs", "outputs"):
            for k, v in self.model_config[n].items():
                if not v["enabled"]:
                    continue
                cmd[k] = format_flag(self.model[n][k]["flag"], fmap[k])

        for k in self.model_config["params"]:
            if self.model["params"][k]["type"] == "bool": ## Handle boolean switches
                if self.model_config["params"][k]["value"]:
                    cmd[k] = (self.model["params"][k]["flag"],)
            else:
                cmd[k] = format_flag(self.model["params"][k]["flag"], str(self.model_config["params"][k]["value"]))

        ## Set the flags in the proper order
        if "order" in self.model:
            order = self.model["order"]
        else:
            order = sorted(list(cmd.keys()))

        acmd = []
        for i in order:
            if i not in cmd:
                continue
            acmd.extend(cmd[i])

        return acmd

    def _make_command(self, flags: List[str]) ->  List[str]:
        """Make the final command.

        :param flags: A list of the flags to be included.
        :returns: The final command, to be passed to :meth:`_run_command_`.
        """
        return flags

    @abstractmethod
    def _run_command(self, cmd: List[str]) -> None:
        """Run the actual command.

        Must raise :class:`DispatchException` if the called process runs into a problem.

        :param cmd: The command to execute.
        """

    @abstractmethod
    def _load_output(self, fmap: Mapping[str, str]) ->  None:
        """Put the output files into the appropriate locations.

        It is this function's responsibility to set self._output_files as the files are created, so
        that they can be removed properly if the command fails.

        :param fmap: The file map.
        """

    def _cleanup(self, error: Exception = None) ->  None:
        """Run successful cleanup.

        This function should conduct all of the cleanup necessary for a successful run. This will
        also be called if the run fails. This means that this function should run checks of its
        own to see what actually needs to be cleaned up.

        :param error: The exception, if one occurred. 
        """
        for f in self._created_files:
            try:
                os.remove(f)
            except:
                pass

        self._created_files = []

    def _cleanup_all(self, error: Exception) ->  None:
        """Cleanup everything.

        This will only be called if an error is raised.
        
        :param error: The exception that occurred.
        """
        self._cleanup(error=error)
        for f in self._output_files:
            try:
                os.remove(f)
            except:
                pass

        self._output_files = []

    def _run(self) -> None:
        """Actual run method.

        This is just to make sure the run method can't be called twice at once.
        """
        logging.info("Processing started")

        try:
            logging.info("Validating model configuration...")
            self.progress(DispatchStage.Validate, 0, 0, "Validating model configuration...")
            self._validate_model_config()

            self.progress(DispatchStage.Validate, 0.5, 0, "Assembling command...")
            fmap = self._make_fmap()
            cmd = self._make_command(self._make_flags(fmap))
            self.progress(DispatchStage.Validate, 1, 0, "Done validation")

            self.progress(DispatchStage.Preprocess, 0, 0, "Starting pre-processing...")
            logging.info("Running pre-processing...")
            self._run_processing(inp=True)
            self.progress(DispatchStage.Preprocess, 1, 0, "Done pre-processing")

            self.progress(DispatchStage.Save, 0, 0, "Saving inputs...")
            self._save_input(fmap)
            self.progress(DispatchStage.Save, 1, 0, "Done saving inputs")

            logging.info("Command: "+' '.join(cmd))

            logging.info("Running model...")
            self.progress(DispatchStage.Run, 0, 0, "Running model...")
            self._run_command(cmd)
            self.progress(DispatchStage.Run, 1, 0, "Model finished")

            self.progress(DispatchStage.Load, 0, 0, "Loading outputs...")
            logging.info("Loading output...")
            self._load_output(fmap)
            self.progress(DispatchStage.Load, 1, 0, "Done loading outputs")

            self.progress(DispatchStage.Postprocess, 0, 0, "Starting post-processing...")
            logging.info("Running post-processing...")
            self._run_processing(inp=False)
            self.progress(DispatchStage.Postprocess, 1, 0, "Done post-processing")

            logging.info("Cleaning up inputs...")
            self._cleanup(error=None)

            logging.info("Processing finished")
        except Exception as e:
            self._cleanup_all(e)
            raise e

    def run(self, model: Mapping, model_config: Mapping, progress: Callable[[DispatchStage, float, float, str], None] = print) ->  None:
        """Run the model.

        This is the entry point

        :param model: The model specification.
        :param model_config: The model configuration.
        :param progress: an optional function accepting a ``DispatchStage``, a float on [0,1] 
                         representing progress in the current stage, a float on [0,1] representing 
                         progress in the current operation, and a string with detailed info.
        """
        ## Ensure only one run is underway
        with self._lock:
            self.model = update_normalize_model(model)[0]
            self.model_config = normalize_model_config(self.model, model_config)
            self.progress = progress

            self._run()



import json
import sys

from .normalize import load_model

if len(sys.argv) not in (2, 3):
    sys.stderr.write("Usage: %s <input file> [output file]\nUpdates and normalizes models.\n" % sys.argv[0])
    sys.exit(1)

res = json.dumps(load_model(open(sys.argv[1], 'r'))[0], indent=4)
if len(sys.argv) == 3:
    open(sys.argv[2], 'w').write(res+'\n')
else:
    print(res)



import io
import gzip
import logging
import os
import queue
import threading
import time

try:
    import sys
    sys.path.insert(0, "/home/ben/Programs/Slicer/lib/Python/lib/python3.6/site-packages/SimpleITK-2.0.0rc2.dev133-py3.6-linux-x86_64.egg")
    import SimpleITK as sitk
except ImportError:
    HAS_SITK = False
else:
    HAS_SITK = True

try:
    import itk
except ImportError:
    HAS_ITK = False
else:
    HAS_SITK = True

if not HAS_ITK and not HAS_SITK:
    logging.warning("SimpleITK and ITK are both unavailable, file conversion will not work")

class KeepAliveIterator:
    """Class for implementing iterators with keep-alives."""
    def __iter__(self):
        empty = False
        while True:
            with self.lock:
                if not self.running and self.queue.empty():
                    self.timer.cancel()
                    break
                if not empty:
                    self.last = time.time()
                empty = False
            try:
                yield self.queue.get(timeout=0.5)
            except queue.Empty:
                empty = True
            except GeneratorExit:
                with self.lock:
                    self.running = False
                    self.timer.cancel()
            
    def mainthread(self):
        for l in self.wrap:
            while True:
                try:
                    self.queue.put(l, timeout=0.5)
                except queue.Full:
                    with self.lock:
                        if not self.running:
                            return
                else:
                    break

            with self.lock:
                if not self.running:
                    return

        with self.lock:
            self.running = False

    def timeoutthread(self):
        with self.lock:
            delta = time.time() - self.last
            if not self.running:
                return
            elif self.queue.empty() and delta >= self.timeout:
                try:
                    self.queue.put(self.alt, block=False)
                except queue.Full:
                    pass
                delta = 0
            self.timer = threading.Timer(self.timeout-delta, self.timeoutthread)
            self.timer.start()


    def __init__(self, wrap, alt="\0\n", qsize=100, timeout=5):
        self.wrap = wrap
        self.queue = queue.Queue(maxsize=qsize)
        self.lock = threading.RLock()
        self.running = True
        self.alt = alt
        self.timeout = timeout
        self.last = time.time()

        self.main = threading.Thread(target=self.mainthread)
        self.main.start()
        self.timer = threading.Timer(timeout, self.timeoutthread)
        self.timer.start()

def guess_filetype(header: bytes, partial=""):
    """Try to guess the filetype.

    :param header: A sensible amount of the header, usually about a kilobyte.
    """

    if len(header) >= 2 and header[:2] == bytes((0x1f, 0x8b)): ## Probably GZIP
        bio = io.BytesIO(header)
        try:
            gz = gzip.GzipFile(fileobj=bio, mode="rb")
            ## First try to read enough for a NII header
            decomp = gz.read(384)
        except EOFError:
            ## Now try to read enough for a NRRD header
            try:
                gz.seek(0)
                decomp = gz.read(4)
            except EOFError:
                ## Now just give up
                return ".gz"+partial
        return guess_filetype(decomp, partial=".gz"+partial)
    elif len(header) >= 4 and header[:4] == b"NRRD": ## Probably NRRD
        return ".nrrd"+partial
    elif len(header) >= 384 and header[:2] == bytes((0x5c, 0x01)) and header[344:348] in (b"n+1\0", "ni1\0"): 
        return ".nii"+partial

    return None

def can_convert(fromext: str, toext: str):
    """Figure out if we can (likely) convert from one extension to another.

    :param fromext: The source extension, including leading period.
    :param toext: The destination extension, including leading period.
    """
    good = (".nii", ".nii.gz", ".nrrd",)

    return (HAS_SITK or HAS_ITK) and fromext.lower() in good and toext.lower() in good

def convert_image(source: str, dest: str):
    """Run the source through ITK/SimpleITK to produce the destination.

    This function is intended to be called after :py:func:`can_convert` to try to actually do the 
    conversion; note that it requires the files to be on disk.

    :param source: The source filename.
    :param dest: The destination filename.
    """
    if HAS_SITK:
        im = sitk.ReadImage(source)
        sitk.WriteImage(im, dest)
    elif HAS_ITK:
        im = itk.imread(source)
        itk.imwrite(im, dest)
    else:
        raise Exception("No conversion library is available!")



"""Module for pre/post-processing."""

from collections.abc import Mapping as CMapping
import logging
from typing import Mapping, Callable, Any, Union, Optional, Sequence

class ROMappingWrapper(Mapping):
    def __init__(self, data):
        self._data = data
        self._wrapped = {}

    def __getitem__(self, key):
        if isinstance(self._data[key], CMapping):
            if key not in self._wrapped or self._wrapped[key]._data != self._data[key]:
                self._wrapped[key] = ROMappingWrapper(self._data[key])
            return self._wrapped[key]
        return self._data[key]

    def __eq__(self, other):
        if not isinstance(other, ROMappingWrapper):
            return False
        return self._data == other._data

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return iter(self._data)

_OPS: Mapping[str, Mapping] = {
    "input": {},
    "output": {},
}

processing_ops = ROMappingWrapper(_OPS)

def register_processing(typ: Union[str, Sequence[str], None], name: str, types: Sequence[str], actions: Optional[Mapping]) ->  Callable:
    """Register a processing operation.

    The format of `actions` is:

    .. code-block:: python

       {
           None: {...}, ## Default action
           "action1": {...}, ## Named action
           ...
       }

    Each sub-dictionary is of the same form as the "inputs" or "outputs" object in a model JSON
    file, e.g.:

    .. code-block:: python

       {
           "param1": {
               "name": "Parameter 1",
               "type": "int",
               "min": 0,
               "max": 10,
               "default": 5
           },
           ...
       }

    with the exception that any pre/post sections are invalid. The type may be "segmentation" or
    "volume", but these MUST NOT be modified by any processing operations.

    :param typ: Is either "input", "output", or None, in which case the processing is considered
                valid for both inputs and outputs.
    :param name: Is the name (not user-friendly) for the action
    :param types: Is the type of inputs/outputs (e.g. "segmentation", "volume") that this operation
                  is valid for
    :param actions: Either None or a dictionary describing the types for each action
    """
    if typ is None:
        typ = tuple(_OPS)
    elif isinstance(typ, str):
        typ = (typ,)
    def wrapper(f):
        for t in typ:
            if name in _OPS[t]:
                logging.warning("Processing function %s/%s already registered" % (t, name))
            _OPS[t][name] = (f, types, actions)

        return f
    return wrapper

def dispatch_processing(op: Mapping, op_config: Mapping, node: Any, node_section: Mapping, model: Mapping, model_config: Mapping, inp: bool = True) ->  None:
    """Dispatch a processing operation.

    :param op: the current operation's section in the model
    :param op_config: the current operation's config section from the model_config
    :param node: the node to operate on (e.g. filename)
    :param node_section: the node's section from the model
    :param model: the entire model specification
    :param model: the entire model config
    :param inp: whether or not this is an input
    """
    if inp:
        typ = "input"
    else:
        typ = "output"

    if op["operation"] not in _OPS[typ]:
        raise KeyError("Unknown processing operation %s/%s" % (typ,op["operation"]))

    _OPS[typ][op["operation"]][0](op, op_config, node, node_section, model, model_config)



#!/usr/bin/env python3

import datetime as dt
import enum
import hashlib
import json
from os import path
import os
import queue
import shutil
import threading
import time
import traceback
from typing import Dict, Any
import uuid

import docker
from flask import jsonify, request, send_file, Response

from ..docker import DispatchDocker
from ..model import load_model

from . import app
from .util import KeepAliveIterator, guess_filetype, can_convert, convert_image

for d in ("MODEL_PATH", "SESSION_PATH"):
    d = app.config[d]
    if not path.isdir(d):
        os.makedirs(d)

__version__ = "1.1.0"

class IndexedQueue:
    """Thread-safe queue with indexing. 

    Uses `threading.RLock` for synchronization.

    The common `timeout` parameter is the same as for `threading.RLock.acquire`. 
    """
    def put(self, item: Any) -> Any:
        """Put an item."""
        with self._lock:
            self._queue.append(item)
            self._count.release()

    def get(self, timeout=None) -> None:
        """Get an item."""
        ret = self._count.acquire(timeout=timeout)
        if not ret:
            raise queue.Empty()

        with self._lock:
            return self._queue.pop(0)

    def index(self, item: Any) -> int:
        """Get the index of an item in the queue.

        Has the same behaviour as `List.index`.
        """
        with self._lock:
            return self._queue.index(item)

    def __init__(self):
        self._queue = []
        self._count = threading.Semaphore(0)
        self._lock = threading.RLock()

class DispatchDockerServer(DispatchDocker):
    def _make_command(self, flags):
        ret = super()._make_command(flags)
        with self.session.lock:
            self.session.container = self.container.id
        return ret

    def _cleanup(self, error=None):
        if self.container is not None and "__logs" in self.model_config:
            with open(self.model_config["__logs"], "wb") as f:
                for chunk in self.container.logs(stream=True):
                    f.write(chunk)

        with self.session.lock:
            print("CLEANUP TIME")
            super()._cleanup()
            self.session.set_state(Session.State.FAILED if error else Session.State.COMPLETE)
            if error:
                self.session.info = repr(error)
            self.session.container = None

    def _run_processing(self, inp):
        ## We don't do any here
        return

    def _on_container_start(self):
        with self.session.lock:
            self.session.state = Session.State.WORKING

    def run(self, session: "Session", *args, **kwargs):
        self.session = session
        super().run(*args, **kwargs)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.session = None

class Model:
    class ModelParam:
        def parse(self, value):
            if self.is_file:
                raise NotImplementedError("Cannot validate files")
            elif self.type == "int":
                minv = self.param["min"] if "min" in self.param else -2147483648
                maxv = self.param["max"] if "max" in self.param else 2147483647

                value = int(value) 
                if value < minv or value > maxv:
                    raise ValueError("Must be between %d and %d" % (minv, maxv))
            elif self.type == "float":
                minv = self.param["min"] if "min" in self.param else -3.40282e+038
                maxv = self.param["max"] if "max" in self.param else 3.40282e+038

                value = float(value)
                if value < minv or value > maxv:
                    raise ValueError("Must be between %d and %d" % (minv, maxv))
            elif self.type == "bool":
                value = bool(value)
            elif self.type == "enum":
                value = self.enum[value]
            elif self.type == "string":
                pass 
            else:
                ## Just in case, we catch this in the constructor
                raise ValueError("Unknown type %s (THIS SHOULD NOT HAPPEN)" % self.type)

            return value

        def __init__(self, id_, param):
            if param["type"] not in ("volume", "segmentation", "int", "float", "bool", "enum", "string",):
                raise ValueError("Unknown type %s for %s" % (param["type"], param["name"]))
            self.id = id_
            self.is_file = param["type"] in ("segmentation", "volume",)
            self.param = param
            self.type = param["type"]
            self.default = param["default"] if "default" in param else None
            self.enum = None
            if self.type == "enum":
                if isinstance(self.param["enum"], list):
                    self.enum = {n: n for n in  self.param["enum"]}
                else:
                    self.enum = self.param["enum"]

    def __init__(self, model: Dict):
        self.name = model["name"]
        self.inputs = {k: Model.ModelParam(k, v) for k, v in model["inputs"].items()}
        self.params = {k: Model.ModelParam(k, v) for k, v in model["params"].items()}
        self.outputs = {k: Model.ModelParam(k, v) for k, v in model["outputs"].items()}
        self.model = model

class Session:
    """Class for an inference session.

    The session's `lock` attribute is an RLock and must be acquired before accessing/modifying any
    of the session's attributes.
    """
    class State(enum.Enum):
        WAITING = "waiting"
        READY = "ready"
        WORKING = "working"
        COMPLETE = "complete"
        FAILED = "failed"

    def set_state(self, state: State):
        """Set the state and update the "last used" time."""
        self.state = state
        self.state_change = dt.datetime.now()

    def cleanup(self):
        """Cleanup the session.

        This removes all files remaining on disk.
        """
        try:
            shutil.rmtree(self.base_dir)
        except:
            pass

    def __init__(self, id_, model: Model, params: Dict):
        self.id = id_
        self.model = model
        self.params = params
        self.lock = threading.RLock()
        self.last = dt.datetime.now()

        ## A mapping of needed items to whether or not they're busy. If an item is in here, it's 
        ## needed
        self.needed = {k: False for k, v in model.inputs.items() if v.is_file and params["inputs"][k]["enabled"]}

        self.state_change = dt.datetime.now()
        self.state = Session.State.WAITING
        self.info = ""

        self.base_dir = path.join(app.config["SESSION_PATH"], id_)
        os.mkdir(self.base_dir)
        os.mkdir(path.join(self.base_dir, "inputs"))
        os.mkdir(path.join(self.base_dir, "outputs"))

        ## Store parameters
        with open(os.path.join(self.base_dir, "params.json"), "w") as f:
            json.dump(params, f)

        ## Setup the filenames
        self.filenames = {}
        for s in ("inputs", "outputs"):
            sv = model.model[s]
            for k, v in sv.items():
                self.filenames[k] = path.join(self.base_dir, s, k+v["extension"])
                self.params[s][k]["value"] = self.filenames[k]
        self.params["__logs"] = path.join(self.base_dir, "logs")
        self.params["__error"] = path.join(self.base_dir, "error")

        self.container = None

class Worker:
    def run(self):
        while self.main.running:
            try:
                ## Timeout every 5 seconds to check if we're supposed to quit
                session = self.main.queue.get(timeout=5)
            except queue.Empty:
                continue

            ## Run the model
            try:
                self.dispatch.run(session, session.model.model, session.params, print)
            except Exception as e:
                with session.lock, open(session.params["__error"], 'w') as f:
                    traceback.print_exc(file=f)
                traceback.print_exc()

    def __init__(self, main: "Main", id_):
        self.main = main
        self.id = id_

        ## TODO: Docker config?
        self.dispatch = DispatchDockerServer()

class Main:
    """Main server object."""
    def handle_models(self):
        """Provide a list of the names of available models."""
        return jsonify(data=tuple(self.models.keys()))

    def handle_models_model_inp_check(self, model, inp):
        """Check if an input header is valid, or can be converted."""
        if model not in self.models:
            return jsonify(errors=[{"detail": "Unknown model %s" % model}]), 404

        cl = request.content_length
        if cl > 4096:
            return jsonify(errors=[{"detail": "Request too large, needs at most 4096"}]), 413

        header = bytes()
        ext = self.models[model].model["inputs"][inp]["extension"]
        checked = False

        for chunk in request.stream:
            header += chunk

        ft = guess_filetype(header)
        convert = app.config["CONVERT_INPUTS"] and ft is not None and ft.lower() != ext.lower() and can_convert(ft, ext)
        print("convert", convert, ft, ext)
        return jsonify(data={
            "filetype": ft,
            "acceptable": (ext not in (".nii", ".nrrd", ".nii.gz")) or not app.config["CHECK_INPUT_FILETYPES"] or (ft is not None and ft.lower() == ext.lower()) or convert,
            "can_convert": convert,
        })

    def handle_models_model(self, model):
        """Provide the model specification for the given model name."""
        if model not in self.models:
            return jsonify(errors=[{"detail": "Unknown model %s" % model}]), 404

        if request.method == "GET":
            return jsonify(data=self.models[model].model)

        ## POST, create a new session
        if not request.is_json:
            return jsonify(errors=[{"detail": "Must pass the model parameters as JSON"}]), 400

        ## Validate and parse the user's parameters
        ## FIXME: This should likely be taken from the regular normalization code
        model = self.models[model]
        params = request.get_json()
        params_val = {"params": {}}
        for section in params_val:
            if section not in params:
                return jsonify(errors=[{"detail": "Missing section %s" % section}]), 400
            params_sec = params[section]
            params_val_sec = params_val[section]
            for name, param in getattr(model, section).items():
                params_val_sec[name] = {}
                if param.is_file:
                    continue
                elif name not in params_sec:
                    if param.default is None:
                        return jsonify(errors=[{"detail": "Missing parameter %s/%s" % (section, name)}]), 400
                    value = param.default
                else:
                    if not isinstance(params_sec[name], dict):
                        return jsonify(errors=[{"detail": "Improper formatting for %s/%s" % (section, name)}]), 400
                    try:
                        value = param.parse(params_sec[name]["value"])
                    except Exception as e:
                        return jsonify(errors=[{"detail": "Invalid value for %s/%s: %s" % (section, name, str(e))}]), 400
                params_val_sec[name]["value"] = value
        
        ## Check which inputs/outputs are enabled
        for section in ("inputs", "outputs"):
            params_val[section] = {n: {"enabled": v["status"] in ("required", "optional")} for n, v in model.model[section].items()}
            if section not in params:
                continue
            for n, v in params[section].items():
                if n not in model.model[section]:
                    return jsonify(errors=[{"detail": "Unknown %s %s" % (section[:-1], n)}])
                if "enabled" in v:
                    enabled = v["enabled"]
                    if not enabled and model.model[section][n]["status"] == "required":
                        return jsonify(errors=[{"detail": "%s %s cannot be disabled" % (section.title()[:-1], n)}])
                    params_val[section][n]["enabled"] = enabled

        ## Now that they're valid, create a new session
        with self.sessions_lock:
            id_ = str(uuid.uuid4())
            while id_ in self.sessions:
                id_ = str(uuid.uuid4())

            self.sessions[id_] = Session(id_, model, params_val)

        ## Return the created session ID
        return jsonify(data={"session_id": id_})

    def handle_sessions_session(self, session):
        """Return information on a session."""
        with self.sessions_lock:
            if session not in self.sessions:
                return jsonify(errors=[{"detail": "Unknown session"}]), 404

            session = self.sessions[session]
            with session.lock:
                data = {
                    "status": str(session.state.value),
                    "needed": list(session.needed.keys()),
                    "info": str(session.info),
                }

            return jsonify(data=data)

    def handle_sessions_session_inputs(self, session, inp):
        """Allow the user to upload input files for the session."""
        with self.sessions_lock:
            if session not in self.sessions:
                return jsonify(errors=[{"detail": "Unknown session"}]), 404

            session = self.sessions[session]

            with session.lock:
                if inp not in session.model.inputs:
                    return jsonify(errors=[{"detail": "Unknown input"}]), 404
                elif session.state != Session.State.WAITING:
                    return jsonify(errors=[{"detail": "Session has already received all input"}]), 400

                if inp not in session.needed:
                    ## Overwrite the old one
                    session.needed[inp] = False 
                if session.needed[inp]:
                    ## Busy, ignore the request 
                    return jsonify(errors=[{"detail": "Resource is already being uploaded"}]), 409
                
                ## Set the busy flag
                session.needed[inp] = True
                session.last = dt.datetime.now()

        try:
            h = hashlib.sha256()

            header = bytes()
            ext = session.model.inputs[inp].param["extension"]
            checked = False

            actual_ext = None
            convert = False
            fname = session.filenames[inp]

            with open(fname, "wb") as f:
                for chunk in request.stream:
                    chunk = bytes(chunk)
                    if not checked:
                        if len(header) < 1024:
                            header += chunk[:1024]
                        if len(header) >= 1024:
                            if app.config["CHECK_INPUT_FILETYPES"] and ext in (".nii", ".nii.gz", ".nrrd"):
                                ft = guess_filetype(header)
                                if ft is not None and ft.lower() != ext.lower():
                                    if app.config["CONVERT_INPUTS"] and can_convert(ext, ft):
                                        ## We'll try conversion later
                                        convert = True
                                        actual_ext = ft
                                    else:
                                        return jsonify(errors=[{"detail": "Expected filetype %s" % ext}]), 400
                            checked = True
                            header = None

                    f.write(chunk)
                    h.update(chunk)

            if convert:
                ## First, move it so the conversion picks up the actual extension
                moved_fname = fname.replace('.', '_') + actual_ext
                os.rename(fname, moved_fname)
                try:
                    convert_image(moved_fname, fname)
                except Exception as e:
                    return jsonify(errors=[{"detail": "Expected filetype %s" % ext}, {"detail": "Attempted conversion failed: "+str(e)}]), 400
                finally:
                    os.remove(moved_fname)

            with session.lock:
                del session.needed[inp]

                if not session.needed:
                    session.state = Session.State.READY 
                    self.queue.put(session)

        finally:
            with session.lock:
                ## Clear the busy flag, if something broke
                if inp in session.needed:
                    session.needed[inp] = False
            
        return jsonify(data={"sha256": h.hexdigest()})

    def handle_sessions_session_outputs(self, session, outp):
        """Allow the user to download finished output files."""
        with self.sessions_lock:
            if session not in self.sessions:
                return jsonify(errors=[{"detail": "Unknown session"}]), 404

            session = self.sessions[session]
            
            with session.lock:
                if outp not in session.model.outputs:
                    return jsonify(errors=[{"detail": "Unknown output"}]), 404
                elif not session.params["outputs"][outp]["enabled"]:
                    return jsonify(errors=[{"detail": "Output was not enabled"}])
                elif session.state != Session.State.COMPLETE:
                    return jsonify(errors=[{"detail": "Session is not complete"}]), 400
            
            return send_file(session.filenames[outp])

    def cleanup(self):
        """Run cleanup of old sessions."""
        print("Running cleanup")
        now = dt.datetime.now()
        remove_states = [Session.State.WAITING]
        if app.config["CLEANUP_FINISHED"]:
            remove_states.extend((Session.State.FAILED, Session.State.COMPLETE,))

        with self.sessions_lock:
            sessions = list(self.sessions.keys())
            for s in sessions:
                v = self.sessions[s]
                with v.lock:
                    if v.state in remove_states and (now - v.last).total_seconds() >= app.config["CLEANUP_SECONDS"]:
                        if app.config["CLEANUP_FILES"]:
                            v.cleanup()
                        del self.sessions[s]

            self.cleanup_timer = threading.Timer(app.config["CLEANUP_TIMER"], self.cleanup)
            self.cleanup_timer.start()

    def handle_sessions_session_logs(self, session):
        """Return session logs."""
        with self.sessions_lock:
            if session not in self.sessions:
                return jsonify(errors=[{"detail": "Unknown session"}]), 404

            session = self.sessions[session]
            with session.lock:
                if session.state not in (Session.State.WORKING, Session.State.READY, Session.State.COMPLETE, Session.State.FAILED):
                    return jsonify(errors=[{"detail": "Session not running yet"}]), 400
                elif session.state == Session.State.READY:
                    container = True
                else:
                    container = session.container

        if container is not None:
            ## Get logs from the container
            def gen():
                while True:
                    with self.sessions_lock, session.lock:
                        if session.state == Session.State.READY:
                            try:
                                index = self.queue.index(session) + 1
                            except ValueError:
                                index = 0
                        elif session.state == Session.State.WORKING:
                            client = docker.from_env()
                            container = client.containers.get(session.container)
                            break
                        elif session.state in (Session.State.FAILED, Session.State.COMPLETE):
                            ## Handle the case where the images finished very quickly
                            with open(session.params["__logs"], "r") as f:
                                for line in f:
                                    yield line.encode("utf-8")
                            return
                    yield ("Waiting, position in queue is %d\n" % index).encode("utf-8")
                    time.sleep(1)

                for line in container.logs(stream=True, follow=True):
                    yield line
            return Response(KeepAliveIterator(gen(), alt="\0\n"), mimetype="text/plain", headers={"X-Accel-Buffering": "no"})
        else:
            return send_file(session.params["__logs"], mimetype="text/plain")

    def handle_root(self):
        return jsonify(data={"server": "inferserver", "version": __version__})

    def __init__(self):
        self.models = {}
        for fname in os.listdir(app.config["MODEL_PATH"]):
            if not fname.lower().endswith(".json"):
                continue
            with open(path.join(app.config["MODEL_PATH"], fname), 'r') as f:
                model = load_model(f)[0]
            self.models[model["id"]] = Model(model)

        self.sessions = {}
        self.sessions_lock = threading.RLock()

        app.route("/", methods=["GET"])(self.handle_root)
        app.route("/models", methods=["GET"])(self.handle_models)
        app.route("/models/<model>", methods=["GET", "POST"])(self.handle_models_model)
        app.route("/models/<model>/inputs/<inp>/check", methods=["POST"])(self.handle_models_model_inp_check)
        app.route("/sessions/<session>", methods=["GET"])(self.handle_sessions_session)
        app.route("/sessions/<session>/inputs/<inp>", methods=["PUT"])(self.handle_sessions_session_inputs)
        app.route("/sessions/<session>/outputs/<outp>", methods=["GET"])(self.handle_sessions_session_outputs)
        app.route("/sessions/<session>/logs", methods=["GET"])(self.handle_sessions_session_logs)

        self.running = True

        self.queue = IndexedQueue()
        self.workers = []

        if app.config["CLEANUP_TIMER"] != -1:
            self.cleanup_timer = threading.Timer(app.config["CLEANUP_TIMER"], self.cleanup)
            self.cleanup_timer.start()
        else:
            self.cleanup_timer = None

        for i in range(app.config["WORKER_THREADS"]):
            worker = Worker(self, i)
            thread = threading.Thread(target=worker.run)
            self.workers.append(thread)
            thread.start()

main = Main()



"""This module provides utility functions for loading, updating, and normalizing model 
specifications. Typically, an application should use only :meth:`load_model` to load files, but the
other functions are also exposed.

Unless stated otherwise, all functions here modify the passed model directly and only return it for
convenience.

This submodule may also be called (e.g. with ``python3 -m ablinfer.model``) to update and normalize
model specification files.
"""

from .update import update_model
from .normalize import load_model, normalize_model, update_normalize_model
from .config import normalize_model_config



import requests as r
import time

class ProgressReporter:
    def _report(self, chunk_length):
        ct = time.monotonic()
        ## Apparently Windows doesn't give enough precision here
        delta = max(ct - self.last, 0.0000001)
        self.avg = 0.8*self.avg + 0.2*(chunk_length/delta)
        self.running += chunk_length
        self.last = ct
        if ct-self.last_report > self.period:
            self.last_report = ct
            r_mib = self.running / 1048576
            t_mib = self.total / 1048576
            a_mib = self.avg / 1048576
            if self.total != -1:
                percent = min(1, max(0, self.running/self.total))
                self.func(percent, self.header + " %.1f/%.1f MiB (%.1f MiB/s)" % (r_mib, t_mib, a_mib))
            else:
                self.func(0, self.header + " %.1f/- MiB (%.1f MiB/s)" % (r_mib, a_mib))

    def report(self, chunk_length):
        self.toadd += chunk_length
        if self.toadd > 512*1024 or self.running+self.toadd == self.total:
            self._report(self.toadd)
            self.toadd = 0

    __call__ = report

    def __init__(self, total, header, func, period = 0):
        self.total = total
        self.running = 0
        self.header = header
        self.func = func
        self.last = time.monotonic()
        self.last_report = self.last
        self.avg = 0
        self.period = period
        self.toadd = 0

        func(0, header)

class FObjReadWrapper:
    def read(self, n=-1):
        buff = self.fp.read(n)
        self.pr(len(buff))

        return buff

    def tell(self):
        return self.fp.tell()

    def seek(self, n):
        return self.fp.seek(n)

    def __len__(self):
        return self.size

    def __init__(self, fp, size, header, func, period = 0):
        self.fp = fp
        self.size = size
        self.pr = ProgressReporter(size, header, func, period)

def save_resp(resp, fname, header="Downloading file...", func=print, period=0.5):
    pr = ProgressReporter(
        int(resp.headers.get("content-length", -1)), 
        header, 
        func, 
        period=period,
    )

    with open(fname, "wb") as f:
        for chunk in resp:
            f.write(chunk)
            pr(len(chunk))



import os
from flask import Flask

app = Flask(__name__)
app.config.from_object("ablinfer.inferserver.default_settings")
if os.environ.get("INFERSERVER_SETTINGS") is not None:
    app.config.from_envvar("INFERSERVER_SETTINGS")

from .views import main



#!/usr/bin/env python3

"""Module for handling model validation and updating.

The main purpose is to ensure that loaded models are valid and that all other functions may assume
that each model conforms to the newest model specification format. Future model specification
formats are indended to be backwards-compatible with old ones, with sane defaults for new fields
added automatically.

Each update function here is intended to update the model from one version to the next. The
functions will then be run in sequence until the model has been brought to the latest version.
This version is determined by the "json_version" field; if absent, the model is assumed to be a
DeepInfer model.

v1.3
- Added "status" field to inputs and outputs to allow optional I/O

v1.2
- Added the "ID" field
- Added the "colours" and "names" parameters to "segmentation" parameters

v1.1
- Added the "website" field and removed "brief_description"

v1.0
- Unified metadata fields from various DeepInfer models
- Standardize certain fields in the various input/output/parameter types
- Remove the plethora of integer types in favour of min/max values
- Add pre/post-processing functions
- Add more description to each field
"""

from collections import OrderedDict as OD
import json
import hashlib
import logging
import re
from typing import Tuple, Callable, IO, Dict, Union

__version__ = "1.3"
__version__int = tuple((int(i) for i in __version__.split('.')))

_UPDATES = {}
def _register(s: str) ->  Callable:
    """Register an update function."""
    def reg_inner(f: Callable[[Dict], Dict]) ->  Callable[[Dict], Dict]:
        if s in _UPDATES:
            raise ValueError("Already registered a helper for v%s" % s)
        _UPDATES[s] = f

        return f
    return reg_inner

@_register("deepinfer")
def _update_deepinfer(model):
    """Update from the DeepInfer model format."""

    ## DeepInfer's regex for converting CamelCase to a friendly name
    re_camel = re.compile('((?<=[a-z0-9])[A-Z]|(?!^)[A-Z](?=[a-z]))')
    beautify_cc = lambda s: re_camel.sub(r' \1', s)

    inor = lambda d, v, s: d[v] if v in d else s

    ## This is a big one, we create the new model from scratch
    nm = OD((
        ("json_version", "1.0"),
        ("type", "docker"),

        ("name", model["name"]),
        ("organ", inor(model, "organ", "")),
        ("task", inor(model, "task", "")),
        ("status", inor(model, "status", "")),
        ("modality", inor(model, "modality", "")),
        ("version", inor(model, "version", "")),

        ("description", inor(model, "detaileddescription", inor(model, "briefdescription", ""))),
        ("brief_description", inor(model, "briefdescription", "")),
        ("maintainers", inor(model, "maintainers", [])),
        ("citation", inor(model, "citation", "")),

        ("docker", OD((
            ("image_name", model["docker"]["dockerhub_repository"]),
            ("image_tag", model["docker"]["digest"]),
            ("data_path", inor(model, "data_path", "/home/deepinfer/data")),
        ))),

        ("inputs", OD()),
        ("params", OD()),
        ("outputs", OD()),
    ))

    for member in model["members"]:
        name = member["name"]
        iotype = member["iotype"]
        typ = member["type"]
        if typ in ("bool", "int",) and iotype != "parameter":
            logging.info("Converting %s to parameter from %s" % (name, iotype))
            iotype = "parameter"

        if iotype in ("input", "output"):
            if typ == "volume":
                ## Note: itk_type and default are both ignored by DeepInfer for volumes
                nm[iotype+'s'][name] = OD((
                    ("name", beautify_cc(name)),
                    ("description", inor(member, "detaileddescriptionSet", "")),
                    ("flag", "--"+name),
                    ("extension", ".nrrd"),

                    ("type", "volume"),
                    ("labelmap", member["voltype"] == "LabelMap"),
                ))
            elif typ == "point_vec":
                nm[iotype+'s'][name] = OD((
                    ("name", beautify_cc(name)),
                    ("description", inor(member, "detaileddescriptionSet", "")),
                    ("flag", "--"+name),
                    ("extension", ".fcvs"),

                    ("type", "point_vec"),
                ))
            else:
                raise ValueError("Unknown %s type %s" % (iotype, typ))
        elif iotype == "parameter":
            m = OD((
                ("name", beautify_cc(name)),
                ("description", inor(member, "detaileddescriptionSet", "")),
                ("flag", "--"+name),

            ))

            rmap = {
                "uint8_t": (0, 255),
                "int8_t":(-128, 127),
                "uint16_t":(0, 65535),
                "int16_t":(-32678, 32767),
                "uint32_t": (0, 2147483647),
                "uint64_t": (0, 2147483647),
                "unsigned int": (0, 2147483647),
                "int32_t": (-2147483648, 2147483647),
                "int64_t": (-2147483648, 2147483647),
                "int": (-2147483648, 2147483647),
            }

            if typ in ("uint8_t", "int8_t", "uint16_t", "int16_t", "uint32_t", "int32_t", "uint64_t", "int64_t", "unsigned int", "int"):
                m["type"] = "int"
                m["default"] = inor(member, "default", 0)
                minv, maxv = rmap[typ]

                m["min"] = minv
                m["max"] = maxv
            elif typ == "bool":
                m["type"] = "bool"
                m["default"] = (inor(member, "default", "false") == "false")
            elif typ in ("float", "double"):
                m["type"] = "float"
                m["default"] = float(inor(member, "default", 0))
            elif typ == "enum":
                m["type"] = "enum"
                m["enum"] = OD(((i, i) for i in member["enum"]))
                m["default"] = inor(member, "default", member["enum"][0])

    return nm

@_register("1.0")
def update_1_0(model):
    model["json_version"] = "1.1"
    if "website" not in model:
        model["website"] = ""
    if "brief_description" in model:
        del model["brief_description"]

    return model

@_register("1.1")
def update_1_1(model):
    model["json_version"] = "1.2"
    if "id" not in model:
        tid = model["name"].lower().replace(' ', '_').replace('-', '_')
        if not tid.isidentifier():
            tid = "model_" + hashlib.md5(model["name"].encode("utf-8")).hexdigest()
        model["id"] = tid

    return model

@_register("1.2")
def update_1_2(model):
    model["json_version"] = "1.3"
    for s in ("inputs", "outputs"):
        for v in model[s].values():
            if "status" not in v:
                v["status"] = "required"

    return model

def update_model(model: Dict, updated: bool = False) ->  Tuple[Dict, bool]:
    """Update a model to the newest version.

    Note that no verification of fields not affected by the updates is conducted.

    :returns: The updated model and a boolean indicating whether an update was performed.
    """
    if "json_version" in model:
        v = model["json_version"]

        try:
            major, minor = (int(i) for i in v.split('.'))
            v_i = (major, minor)
        except:
            raise ValueError("Model version %s is invalid (should be major.minor)" % v)

        ## Normalize the version, just in case
        v = "%d.%d" % v_i

        if v_i > __version__int:
            raise ValueError("Model version %s is too new (this version is %s)" % (v, __version__))
        elif v == __version__:
            return model, updated
    else:
        v = "deepinfer"

    if v not in _UPDATES:
        raise ValueError("I don't know how to update v%s" % v)

    nm = _UPDATES[v](model)

    return update_model(nm)[0], True



"""Docker helper to simplify retrieving/sending single files."""

import os
import shutil
import tarfile

import docker

class _DLReader:
    """Wrapper to turn an iterator into a file object.

    This is for piping the result of a streaming `requests.get` into `tarfile`.
    """
    def read(self, count):
        buff = self.last
        while len(buff) < count:
            try:
                buff += next(self.it)
            except StopIteration:
                break
        if len(buff) > count:
            self.last = buff[count:]
            buff = buff[:count]
        else:
            self.last = bytes()

        self.offset += len(buff)

        return buff

    def seekable(self):
        return False

    def tell(self):
        return self.offset

    def seek(self, offset):
        todo = offset - self.offset
        if not todo:
            return
        elif todo < 0:
            raise ValueError("Can't seek backwards")

        buff = bytes()
        while todo > 0:
            try:
                buff = next(self.it)
            except StopIteration:
                break
            self.offset += len(buff)
            todo -= len(buff)
        if todo < 0:
            self.last = buff[todo:]

    def __init__(self, it):
        self.it = it
        self.last = bytes()
        self.offset = 0

class _SingleFileTar:
    """Quick-and-dirty tar creation implementation.

    Intended to avoid using a temporary archive file when the created archive just needs to be
    read as a file into some other function. Only works for regular files and doesn't store any
    metadata besides the filename.
    """
    def read(self, count):
        if self.offset == self.size:
            return b""

        while len(self.buff) < count:
            buff = self.fp.read(self.buffsize)
            self.buff += buff
            if len(buff) < self.buffsize: ## We've run out of file add the padding
                self.buff += b'\0'*(self.pad + 1024)
                break

        count = min(count, len(self.buff))

        if self.offset+count > self.size:
            print(self.offset, count, self.size)
            raise ValueError()
        self.offset += count
        ret = self.buff[:count]
        self.buff = self.buff[count:]

        return ret

    def close(self):
        self.fp.close()
        self.fp = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __iter__(self):
        assert self.offset == 0
        return self

    def __next__(self):
        buff = self.read(self.buffsize)

        if not buff:
            raise StopIteration

        return buff

    def __init__(self, fname, name=None, buffsize=2097152, mode=0o640):
        self.offset = 0
        self.fname = fname

        self.buffsize = buffsize

        self.fp = open(self.fname, "rb")

        ## Construct the header
        size = os.path.getsize(fname)

        header = bytes()

        pad = lambda i, s, p=b'\0': i + (p * max(s-len(i), 0))
        pad_num = lambda i, s, p='0': ((p*max(s-len(i)-1, 0)) + i).encode("ascii") + b'\0'

        ## File name
        header += pad((os.path.basename(fname) if name is None else name).encode("ascii")+b'\0', 100)
        ## File mode (ignore for now)
        header += pad_num(oct(mode)[2:][-3:], 8)
        ## File owner (numeric, as ASCII)
        header += pad_num("", 8)
        ## File group (numeric, as ASCII)
        header += pad_num("", 8)
        ## File size in bytes, as an octal string, padded with zeros
        header += pad_num(oct(size)[2:], 12)
        ## Modification time, Unix format, in octal
        header += pad(b"", 12)
        ## Checksum for header, MUST BE SPACES WHEN CALCULATING IT
        header += b' '*8
        ## File type
        header += b'\0'
        ## Link name
        header += pad(b"", 100)
        ## Ignore UStar fields, we don't care
        header = pad(header, 512)

        ## Calculate and replace the checksum
        checksum = sum(header)
        ## 6-digit octal number padded with zeros if necessary, plus a NUL and then a space (???)
        header = header[:148] + pad_num(oct(checksum)[2:], 6) + b"\0 " + header[156:]

        self.buff = header

        ## Total number of blocks: the header is one, plus the file padded to the nearest block,
        ## plus the final two empty blocks (so +3)
        blocks = (size + 511)//512 + 3

        if size%512:
            self.pad = 512 - size%512
        else:
            self.pad = 0

        self.size = blocks * 512

def put_file(
        container: docker.models.containers.Container, 
        path: str, 
        fname: str,
        name: str = None,
        mode: int = 0o640,
    ) ->  None:
    """Put a single file into a container.

    Only works on single regular files and ignores all metadata.

    :param path: The directory in the container to extract to.
    :param fname: The filename of the local file.
    :param name: The name to store in the tar file; defaults to the basename of the file.
    :param mode: The mode for the stored file (3-digit octal number).
    """

    with _SingleFileTar(fname, name, mode=mode) as f:
        container.put_archive(path, f)

def get_file(container: docker.models.containers.Container, path: str, dest: str = ".") ->  None:
    """Get a single file from a container.

    :param path: The path to retrieve.
    :param dest: The filename to extract to. If it is an existing directory, the filename from the
                 tar file will be used.
    """
    if os.path.isdir(dest):
        fname = None
    else:
        fname = dest

    it, _ = container.get_archive(path)
    reader = _DLReader(it)
    tf = tarfile.TarFile(mode="r", fileobj=reader)
    for member in tf:
        ## We only need the first one
        f = tf.extractfile(member)
        if fname is None:
            fname = os.path.join(dest, member.name)

        with open(fname, "wb") as of:
            shutil.copyfileobj(f, of)



#!/usr/bin/env python3

from collections.abc import Collection as ABCCollection
from collections.abc import Mapping as ABCMapping
from collections import OrderedDict as OD
import json
import logging
from numbers import Number
from typing import Mapping, Optional, Collection, Union, IO, Tuple, Dict, Set, Any
from .typing_shim import get_origin, get_args
import collections
from .update import __version__, update_model
from ..processing import processing_ops

def get_origin_w(tp):
    o = get_origin(tp)
    return o if o is not None else tp

def normalize_model(model: Mapping, processing: bool = False) -> Mapping:
    """Normalize the given model. 

    Also validates what fields it can on the model.

    :param model: The model to normalize.
    :param processing: Whether or not to parse in detail the processing operations.
    """
    if model["json_version"] != __version__:
        raise ValueError("Can only normalize v%s models, please update it first" % __version__)

    def isinstance_w(a, t):
        if isinstance(t, Collection):
            if Any in t:
                return True
            return isinstance(a, t)
        if t == Any:
            return True
        return isinstance(a, t)

    def check_rec(v, t):
        origin = get_origin_w(t)
        args = get_args(t)
        if origin == Union:
            if None in args and v is None:
                return True
            for a in args:
                if a is None:
                    continue
                ret = check_rec(v, a)
                if ret:
                    return True
            return False
            #return any((check_rec(v, a) for a in args if a is not None))
        elif origin in (Collection, ABCCollection):
            if args:
                return isinstance(v, ABCCollection) and all((check_rec(i, args[0]) for i in v))
            return isinstance(v, ABCCollection) 
        elif origin in (Mapping, ABCMapping):
            if args:
                return isinstance(v, ABCMapping) and all((isinstance_w(i, args[0]) and isinstance_w(j, args[1]) for i, j in v.items()))
            return isinstance(v, Mapping)
        return isinstance_w(v, origin)

    def simple_verify(d, spec_part, fragment=(), warn_extra=True):
        if warn_extra:
            for k in d:
                if k not in spec_part:
                    logging.warning("Extraneous field %s" % '/'.join(fragment+(k,)))
        for k, t in spec_part.items():
            if isinstance(t, Collection) and not isinstance(t, Mapping):
                t, c = t
            else:
                c = None

            optional = (get_origin_w(t) == Union) and (type(None) in t.__args__)

            if not optional and k not in d:
                raise ValueError("Missing required field %s" % '/'.join(fragment+(k,)))
            elif optional and k not in d:
                logging.warning("Missing optional field %s" % '/'.join(fragment+(k,)))
                if callable(c):
                    d[k] = c()
                elif c is not None:
                    d[k] = c
            elif isinstance(t, Mapping): ## Recurse
                simple_verify(d[k], t, fragment=fragment+(k,), warn_extra=warn_extra)
            elif not check_rec(d[k], t):
                raise ValueError("Improper type for %s" % '/'.join(fragment+(k,)))

    ## All of these specs indicate the structure of the object and have the following rules
    ## - Each string name maps to a type, a duple, or a mapping
    ## - If mapped to a mapping, then the verification recurses
    ## - If mapped to a type, the value is checked to see if it's that type
    ## - If mapped to a type and the type is Optional[...], then a missing value only triggers a 
    ##   warning, not an exception
    ## - If mapped to a tuple, the first element is the type and the second is what should be used 
    ##   to populate the value if it's missing (only used if the type is Optiona[...])
    ## - If mapped to a tuple, the second element is either a callable, a boolean, or None; in the 
    ##   latter two cases, missing values are set to the given literal and in the former they are 
    ##   set to the result of the callable (with no arguments)

    ## Check the outer layer
    spec = {
        "json_version": str,
        "id": str,
        "type": str,
        "name": str,
        "organ": str,
        "task": str,
        "status": str,
        "modality": str,
        "version": str,
        "description": Optional[str],
        "website": Optional[str],
        "maintainers": (Optional[Collection[str]], list),
        "citation": Optional[str],
        "docker": {
            "image_name": str,
            "image_tag": str,
            "data_path": str,
        },
        "inputs": ABCMapping,
        "params": (Optional[ABCMapping], OD),
        "outputs": ABCMapping,
        "order": Optional[ABCCollection],
    }
    simple_verify(model, spec)

    if not model["id"].isidentifier():
        raise ValueError("Model ID must be a valid Python identifier")

    params: Set[str] = set()
    part_spec = {
        "name": str,
        "description": str,
        "status": str,
        "flag": str,
        "extension": str,
        "type": str,
        "pre": (Optional[ABCCollection], list),
    }

    type_spec = {
        "volume": {
            "labelmap": (Optional[bool], False),
        },
        "segmentation": {
            "labelmap": (Optional[bool], False),
            "master": Optional[str],
            "colours": (Optional[Mapping], OD),
            "names": (Optional[Mapping], OD),
        },
        "int": {
            "min": (Optional[Number], -2147483648),
            "max": (Optional[Number], 2147483647),
            "default": Number,
        },
        "float": {
            "min": (Optional[Number], -3.40282e+038),
            "max": (Optional[Number], 3.40282e+038),
            "default": Number,
        },
        "bool": {
            "default": bool,
        },
        "string": {
            "default": str,
        },
        "enum": {
            "enum": Union[Collection[str], Mapping],
            "default": str,
        },
    }

    process_spec = {
        "name": str,
        "description": str,
        "status": str,
        "locked": (Optional[bool], False),
        "operation": str,
        "action": Optional[str],
        "targets": Optional[Collection[int]],
        "params": Mapping,
    }

    def verify_part(name):
        cname = name.title()
        for k, v in model[name].items():
            if k in params:
                raise ValueError("Names must be unique (%s is already used)" % repr(k))
            elif not isinstance(k, str):
                raise ValueError("%s name %s must be a string" % (cname, repr(k)))
            elif not k.isidentifier():
                raise ValueError("%s name %s must be a valid Python identifier" % (cname, repr(k)))
            params.add(k)
            simple_verify(v, part_spec, fragment=(name, k), warn_extra=False)

            typ = v["type"]
            if name == "params" and typ not in ("int", "float", "bool", "enum", "string"):
                raise ValueError("Invalid %s type %s" % (name, repr(typ)))
            elif name in ("outputs", "inputs") and typ not in ("volume", "segmentation"):
                raise ValueError("Invalid %s type %s" % (name, repr(typ)))
            simple_verify(v, type_spec[typ], fragment=(name, k), warn_extra=False)
            if typ == "enum":
                if not isinstance(v["enum"], Mapping):
                    v["enum"] = OD(((i, i) for i in v["enum"]))
            elif typ == "segmentation":
                if "colours" not in v:
                    v["colours"] = OD()
                for colourk, colourv in v["colours"].items():
                    if not isinstance(colourv, Collection) or len(colourv) not in (3, 4):
                        raise ValueError("Segmentation colours must be 3- or 4-element arrays of floats, not \"%s\"" % repr(colourv))
                    elif len(colourv) == 3:
                        ## Fill in opacity
                        v["colours"][colourk] = (*colourv, 1)
                    for colour in colourv:
                        if colour < 0 or colour > 1:
                            raise ValueError("Invalid segementation colour %s, must be a float on [0,1]" % repr(colour))
            if name in ("outputs", "inputs"):
                if "status" not in v:
                    v["status"] = "required"
                elif v["status"] not in ("required", "suggested", "optional"):
                    raise ValueError("Invalid status %s for %s" % (repr(v["status"]), '/'.join((name, k))))

            if name != "params":
                sname = "pre" if name == "inputs" else "post"
                if sname not in v:
                    v[sname] = []
                for n, sec in enumerate(v[sname]):
                    process_spec_this = process_spec.copy()
                    if sec["operation"] in processing_ops[name[:-1]]:
                        psec = processing_ops[name[:-1]][sec["operation"]][2]
                        if psec is not None:
                            action = sec["action"] if "action" in sec else None
                            if action in psec or None in psec:
                                if action in psec:
                                    act = psec[action]
                                else:
                                    logging.info("Unknown %sprocessing action %s for %s, using default" % (sname, action, sec["operation"]))
                                    act = psec[None]

                                if act:
                                    process_spec_this["params"] = {param_name: (Optional[type_spec[param_val["type"]]["default"]], param_val["default"]) for param_name, param_val in act.items()}
                            else:
                                logging.warning("Unknown %sprocessing action %s for %s, skipping" % (sname, action, sec["operation"]))
                    else:
                        logging.warning("Unknown %sprocessing operation %s, skipping" % (sname, sec["operation"]))
                    simple_verify(sec, process_spec_this, fragment=(name, k, sname, str(n)))
                    if sec["status"] not in ("required", "suggested", "optional"):
                        raise ValueError("Invalid status %s for %s" % (repr(sec["status"]), '/'.join((name, k, sname, str(n)))))

    verify_part("inputs")

    del part_spec["pre"]
    part_spec["post"] = (Optional[Collection], list)
    verify_part("outputs")

    del part_spec["post"]
    del part_spec["status"]
    del part_spec["extension"]
    verify_part("params")

    field_set = set((*model["inputs"], *model["params"], *model["outputs"]))
    flag_set = set((
        *(i["flag"] for i in model["inputs"].values()), 
        *(i["flag"] for i in model["params"].values()),
        *(i["flag"] for i in model["outputs"].values()),
    ))
    if "order" in model and model["order"]:
        if len(model["order"]) != len(model["inputs"]) + len(model["params"]) + len(model["outputs"]):
            raise ValueError("Length of order does not match the number of inputs, outputs, and params")
        elif set(model["order"]) != field_set:
            raise ValueError("Order elements must match field names")
    elif ("order" not in model or not model["order"]) and "" in flag_set:
        raise ValueError("Order must be specified if positional arguments are used")


    return model

def update_normalize_model(model: Mapping) -> Tuple[Mapping, bool]:
    """Update and normalize a model.

    :param model: The model to fix.

    :returns: The updated model and a boolean indicating whether an update was performed.
    """
    upd = update_model(model)
    return (normalize_model(upd[0]), upd[1])

def load_model(f: Union[str, IO], fp: bool = True, normalize: bool = True) ->  Tuple[Mapping, bool]:
    """Load a model.

    :param f: The file or string to load from.
    :param fp: Whether f is a file-like object or a string.
    :param normalize: Whether or not to normalize the result.

    :returns: The updated model and a boolean indicating whether an update was performed.
    """
    if fp:
        inp = json.load(f, object_pairs_hook=OD)
    else:
        inp = json.loads(f, object_pairs_hook=OD)

    upd = update_model(inp)
    if normalize:
        upd = (normalize_model(upd[0]), upd[1])

    return upd



"""Basic implementation for dispatching an inference to a Docker container."""

import logging

import docker
from docker.types import DeviceRequest

from ..base import DispatchBase, DispatchException
from ..constants import DispatchStage
## FIXME: The following import should be removed when DeviceRequest support is added to `docker-py`
from .docker_helper import put_file, get_file

class DispatchDocker(DispatchBase):
    """Class for dispatching to a Docker container.

    A ``docker`` key is added to ``config``, which should contain all of the keyword arguments to
    pass to ``docker.DockerClient``, excepting ``version``. If not present, ``docker.from_env`` is
    used (so either ``docker`` should be present or you should properly set environment variables, 
    the latter being preferred). 

    This implementation expects the model configuration values to be the path to the appropriate 
    files on the **host machine**, which it will then put into/get from the container during the 
    run.
    """
    def __init__(self, config=None):
        self.client = None
        self.container = None

        super(DispatchDocker, self).__init__(config=config)

    def _on_container_start(self):
        """Called when the container is actually started."""
        return

    def _validate_config(self):
        super()._validate_config()

        ## The version here must be "auto". At the time of this writing, `docker-py` defaults to
        ## using a version of the API too old to allow GPU support, even if the server supports it.
        ## Setting it to "auto" negotiates the version to the highest common version.
        if "docker" not in self.config or self.config["docker"] is None:
            self.client = docker.from_env(version="auto")
        else:
            self.client = docker.DockerClient(version="auto", **self.config["docker"])

        self.client.ping()

    def _validate_model_config(self):
        super()._validate_model_config()

        imagename = self.model["docker"]["image_name"] + ':' + self.model["docker"]["image_tag"]
        try:
            self.client.images.get(imagename)
        except docker.errors.ImageNotFound:
            self.progress(DispatchStage.Validate, 0, 0, "Docker image not found locally, trying to pull it")
            self.client.images.pull(imagename)
            self.client.images.get(imagename)

    def _make_fmap(self):
        return self._make_fmap_helper(self.model["docker"]["data_path"])

    def _make_command(self, flags):
        cmd = super()._make_command(flags)

        ## If the API is new enough, request GPU access
        kwargs = {}
        if self.client.version()["ApiVersion"] >= "1.40":
            logging.info("API version is high enough for GPU access")
            ## Request all GPU devices
            kwargs["device_requests"] = [
                {
                    "count": -1,
                    "capabilities": [["gpu"]]
                }
            ]
        else:
            logging.warning("API version is too low for GPUs")

        ## We have to create the container here, since the command/flags need to be known ahead of
        ## time for some strange reason (Docker limitation)
        imagename = self.model["docker"]["image_name"] + ':' + self.model["docker"]["image_tag"]
        self.container = self.client.containers.create(imagename, command=cmd, **kwargs)

        return []

    def _save_input(self, fmap):
        ## We assume here that all of the inputs and outputs are strings, indicating the file
        ## locations on the local machine. We need to copy the input files into the container
        total = len(self.model_config["inputs"])
        for n, (k, v) in enumerate(self.model_config["inputs"].items()):
            if not v["enabled"]:
                logging.info("Skipping disabled input %s" % k)
                continue
            fname, fpath = (i[::-1] for i in fmap[k][::-1].split('/', 1))

            logging.info("Storing file %s to container as %s" % (v["value"], fpath+'/'+fname))
            self.progress(DispatchStage.Save, n/total, 0, "Storing file %s..." % v["value"])
            put_file(self.container, fpath, v["value"], name=fname)

    def _run_command(self, cmd):
        ## Ignore cmd, it's not actually helpful anymore
        self.container.start()
        self._on_container_start()
        for line in self.container.logs(stream=True):
            if isinstance(line, bytes):
                line = line.decode("utf-8")
            self.progress(DispatchStage.Run, 0, 0, line)
        resp = self.container.wait()
        if resp["StatusCode"] != 0:
            ## Don't bother getting stderr, they can just get the logs themselves
            raise DispatchException("Called process failed")

    def _cleanup(self, error=None):
        super()._cleanup(error=error)
        if self.container is not None:
            self.container.remove()

    def _load_output(self, fmap):
        total = len(self.model_config["outputs"])
        for n, (k, v) in enumerate(self.model_config["outputs"].items()):
            if not v["enabled"]:
                logging.info("Skipping disabled output %s" % k)
                continue
            self._output_files.append(v["value"])
            self.progress(DispatchStage.Load, n/total, 0, "Storing file %s...")
            get_file(self.container, fmap[k], v["value"])



"""Module for integrating ABLInfer into Slicer."""

import logging

import os
import slicer

from .. import DispatchBase
from ..docker import DispatchDocker
from ..remote import DispatchRemote

from .processing import __name__ as _

class SlicerDispatchMixin(DispatchBase):
    """Mixin for dispatching from Slicer.

    A ``tmp_path`` key is added to ``config``, which must contain the location to store temporary
    files for dispatching.

    This class does not implement any actual dispatching; this must be combined with the 
    appropriate dispatcher to function properly, e.g.::

        class SlicerDispatchActual(SlicerDispatchMixin, DispatchActual):
            pass
    """
    def __init__(self, config):
        self.tmp_path = None
        self._input_nodes = {}

        super().__init__(config)

    def _validate_config(self):
        super()._validate_config()

        self.tmp_path = self.config["tmp_path"]
        if not os.path.isdir(self.tmp_path):
            os.makedirs(self.tmp_path)

    @staticmethod
    def _clone(a, b=None):
        """Try to clone a node.

        This is copied from qSlicerSubjectHierarchyModuleLogic.cxx, I can't figure out how to call 
        that directly and it doesn't support cloning into another node anyways.

        @param a the original node 
        @param b the node to clone into; if None, creates one
        @returns the created node
        """
        ## TODO: Move this in to an external library?
        if b is None:
            b = slicer.mrmlScene.AddNewNodeByClass(a.GetClassName())
            name = a.GetName()
        else:
            name = b.GetName()

        ## Clone the display node
        a_dn = a.GetDisplayNode()
        if b.GetDisplayNode():
            b_dn = b.GetDisplayNode()
        else:
            b_dn = slicer.mrmlScene.AddNewNodeByClass(a_dn.GetClassName())
            b_dn.Copy(a_dn)
            b_dn.SetName(name + "_Display")
            b.SetAndObserveDisplayNodeID(b_dn.GetID())

        ## Clone storage node
        a_sn = a.GetStorageNode()
        if a_sn:
            if b.GetStorageNode():
                b_sn = b.GetStorageNode()
            else:
                b_sn = slicer.mrmlScene.AddNewNodeByClass(a_sn.GetClassName())
                b_sn.Copy(a_sn)
                if a_sn.GetFileName():
                    b_sn.SetFileName(a_sn.GetFileName())
                b.SetAndObserveStorageNodeID(b_sn.GetID())

        ## Finally, do the copy
        b.Copy(a)
        b.SetName(name)

        b.SetAndObserveDisplayNodeID(b_dn)
        b.SetAndObserveStorageNodeID(b_sn)

        ## Trigger update
        b_ptn = b.GetParentTransformNode()
        if b_ptn:
            b_ptn.Modified()

        return b

    def _save_input(self, fmap):
        self._input_nodes = {}

        for k, v in self.model["inputs"].items():
            if not self.model_config["inputs"][k]["enabled"]:
                continue
            ## Don't use os.path.join here: the Docker container might not have the same OS as the 
            ## host machine, which is where we are now. Forward slashes should work on any system,
            ## so just use them here; we already removed any trailing slash from actual_path
            lpath = os.path.join(self.tmp_path, k+v["extension"])
            ## Write it to the path on the local machine
            if k in self._pre_nodes:
                actual_node = self._pre_nodes[k]
            else:
                actual_node = self.model_config["inputs"][k]["value"]
            ret = slicer.util.saveNode(actual_node, lpath)
            self._created_files.append(lpath)

            self._input_nodes[k] = self.model_config["inputs"][k]["value"]
            self.model_config["inputs"][k]["value"] = lpath

            if not ret:
                raise Exception("Unable to save input \"%s\" to file. THIS SHOULD NOT HAPPEN" % v["name"])

        ## Now let the actual dispatcher put the files into the container
        super()._save_input(fmap)

    def _load_output(self, fmap):
        output_nodes = {}
        ## Point the dispatcher to the local files
        for k, v in self.model_config["outputs"].items():
            output_nodes[k] = v.copy()
            v["value"] = os.path.join(self.tmp_path, k+self.model["outputs"][k]["extension"])

        ## Load them to the local disk
        super()._load_output(fmap=fmap)

        ## Restore the input nodes
        for k, v in self._input_nodes.items():
            if not self.model_config["inputs"][k]["enabled"]:
                continue
            self.model_config["inputs"][k]["value"] = v

        ## Now load them into Slicer
        for k, member in self.model["outputs"].items():
            if not self.model_config["outputs"][k]["enabled"]:
                continue
            logging.info("Loading \"%s\"..." % member["name"])
            of = self.model_config["outputs"][k]["value"]
            if member["type"] == "segmentation":
                if member["labelmap"]:
                    logging.info("- Loading as LabelVolume")
                    _, lvnode = slicer.util.loadLabelVolume(of, returnNode=True)
                    if lvnode is None:
                        raise Exception("Missing output %s! Please report this." % member["name"])
                    node = slicer.mrmlScene.AddNewNodeByClass("vtkMRMLSegmentationNode")
                    logging.info("- Converting to segmentation")
                    slicer.vtkSlicerSegmentationsModuleLogic.ImportLabelmapToSegmentationNode(lvnode, node, "")
                    slicer.mrmlScene.RemoveNode(lvnode)
                else:
                    _, node = slicer.util.loadSegmentation(of, returnNode=True)

                ## Fill in the colours and names

                ## First, map the label values to the actual segment objects
                segmap = {}
                segmentation = node.GetSegmentation()
                display_node = node.GetDisplayNode()
                if display_node is None:
                    logging.warning("Can't find display node for segmentation, opacity will be skipped")
                for i in range(segmentation.GetNumberOfSegments()):
                    thisseg, thisid = segmentation.GetNthSegment(i), segmentation.GetNthSegmentID(i)
                    segmap[thisseg.GetLabelValue()] = (thisseg, thisid)

                ## Now set the names and colours
                for label in set(member["colours"]).union(member["names"]):
                    try:
                        ilabel = int(label)
                    except ValueError:
                        logging.warning("Invalid label %s, ignoring" % (repr(label)))
                        continue
                    if ilabel not in segmap:
                        logging.warning("Couldn't find segment matching label %d, ignoring" % label)
                        continue
                    theseg, theid = segmap[ilabel]
                    if label in member["colours"]:
                        thecolour = member["colours"][label]
                        theseg.SetColor(tuple(thecolour[:3]))
                        if len(thecolour) == 4: ## Opacity
                            display_node.SetSegmentOpacity3D(theid, thecolour[3])
                    if label in member["names"]:
                        theseg.SetName(member["names"][label])

            elif member["type"] == "volume":
                if member["labelmap"]:
                    _, node = slicer.util.loadLabelVolume(of, returnNode=True)
                else:
                    _, node = slicer.util.loadVolume(of, returnNode=True)
            else:
                raise Exception("Unknown output type %s" % repr(member["type"]))

            ## Make sure the output is where it should be
            if output_nodes[k]["value"] is not None:
                ## Clone it
                self.clone(node, output_nodes[k]["value"])
                slicer.mrmlScene.RemoveNode(node)
                node = output_nodes[k]["value"]

            self.model_config["outputs"][k]["value"] = node

    def _cleanup(self, error=None):
        ## Get rid of output too
        self._created_files.extend(self._output_files)
        super()._cleanup(error=error)
        
        ## Remove the pre-processing clones
        for v in self._pre_nodes.values():
            slicer.mrmlScene.RemoveNode(v)
        self._pre_nodes = {}
        self._input_nodes = {}

class SlicerDispatchDocker(SlicerDispatchMixin, DispatchDocker):
    """Convenience class for dispatching to Docker from Slicer."""
    pass

class SlicerDispatchRemote(SlicerDispatchMixin, DispatchRemote):
    """Convenience class for dispatching to an ABLInfer server from Slicer."""
    pass



import setuptools

setuptools.setup(
    name="ablinfer",
    version="1.2.2",
    author="Ben Connors",
    description="Library for dispatching medical images to registration and segmentation toolkits.",
    url="https://github.com/Auditory-Biophysics-Lab/ablinfer",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.3",
    install_requires=[
        "docker>=4.3.0",
        "requests",
    ],
)



"""Module for pre/post-processing."""

import logging

import slicer 

from ..processing import register_processing

@register_processing("output", "render_3d", ["segmentation"], {
    None: {
        "smoothing": {
            "name": "Surface Smoothing Factor",
            "type": "float",
            "min": 0,
            "max": 1,
            "default": 0.5,
        }
    }
})
def render_3d(op, op_config, node, *args, **kwargs):
    name = slicer.vtkBinaryLabelmapToClosedSurfaceConversionRule.GetSmoothingFactorParameterName()
    node.GetSegmentation().SetConversionParameter(name, str(op_config["params"]["smoothing"]))
    node.CreateClosedSurfaceRepresentation()

@register_processing(None, "seged", ["segmentation"], None)
def seged(op, op_config, node, node_config, model, model_config):
    """Segment editor processing."""
    sew = slicer.qMRMLSegmentEditorWidget()
    sew.setMRMLScene(slicer.mrmlScene)
    sen = slicer.mrmlScene.AddNewNodeByClass("vtkMRMLSegmentEditorNode")
    sew.setMRMLSegmentEditorNode(sen)
    segmentation = node.GetSegmentation()
    sew.setSegmentationNode(node)
    sew.setActiveEffectByName(op["action"])

    if not op["targets"]:
        op["targets"] = range(segmentation.GetNumberOfSegments())
    
    if "master" in node_config:
        mv = model_config["inputs"][node_config["master"]]["value"]
    else:
        mv = None
    sew.setMasterVolumeNode(mv)

    for n in op["targets"]:
        logging.info("- Applying to segment %d" % n)

        sew.setCurrentSegmentID(segmentation.GetNthSegmentID(n))
        effect = sew.activeEffect()
        if effect is None:
           raise Exception("Unknown seged action: %s" % op["action"])

        for p, v in op_config["params"].items():
            effect.setParameter(p, str(v))
        
        effect.self().onApply()
    
    sew.setActiveEffectByName(None)
    slicer.mrmlScene.RemoveNode(sen)



MODEL_PATH = "/etc/inferserver/models" ## Directory to look for model JSON files in
SESSION_PATH = "/var/inferserver"      ## Directory to store sessions in *MUST BE WRITABLE*
CLEANUP_TIMER = 300                    ## Timer in between running session cleanup, in seconds
CLEANUP_FINISHED = True                ## Cleanup finished inferences?
CLEANUP_FILES = False                  ## Remove cleaned sessions from the disk?
CLEANUP_SECONDS = 3600                 ## Number of seconds before a session is considered inactive
CHECK_INPUT_FILETYPES = True           ## Verify input file types before dispatching
WORKER_THREADS = 1                     ## Number of simultaneous inferences that can be run
CONVERT_INPUTS = True                  ## Whether or not to try converting inputs



"""This module provides functions and classes for interfacing with 3DSlicer.

When running in 3DSlicer, this module should be explictly imported to access the custom pre/post-
processing functions it provides.
"""

from .dispatchslicer import SlicerDispatchDocker, SlicerDispatchMixin, SlicerDispatchRemote
from .processing import __name__ as _
